from getpass import getpass

from tatu.okast import OkaSt, j


def gettoken(url, username=None, password=None):
    tries = 0
    ntries = 3
    data = {"username": username, "password": password} if username else None

    while tries < ntries:
        tries += 1
        data = data or {
            "username": input("Username to connect to OKA: "),
            "password": getpass("Password to connect to OKA: ")
        }
        okast = OkaSt(token=None, url=url, close_when_idle=True)
        response = okast.request(f"/api/auth/login", "post", json=data)
        if response and 'access_token' in j(response):
            return j(response)['access_token']
        else:
            if tries < ntries:
                print("[Error] Authentication error. Please try again.")
        data = None
    print("[Error] Authentication failed.")
    exit()



# #  Copyright (c) 2020. Davi Pereira dos Santos
# #  This file is part of the tatu project.
# #  Please respect the license - more about this in the section (*) below.
# #
# #  tatu is free software: you can redistribute it and/or modify
# #  it under the terms of the GNU General Public License as published by
# #  the Free Software Foundation, either version 3 of the License, or
# #  (at your option) any later version.
# #
# #  tatu is distributed in the hope that it will be useful,
# #  but WITHOUT ANY WARRANTY; without even the implied warranty of
# #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# #  GNU General Public License for more details.
# #
# #  You should have received a copy of the GNU General Public License
# #  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
# #
# #  (*) Removing authorship by any means, e.g. by distribution of derived
# #  works or verbatim, obfuscated, compiled or rewritten versions of any
# #  part of this work is a crime and is unethical regarding the effort and
# #  time spent here.
# #  Relevant employers or funding agencies will be notified accordingly.
#
# import socket
#
# import pymysql
# import pymysql.cursors
# from pymysql.constants import CLIENT
# from sqlalchemy.engine import Engine
#
# from garoupa.decorator import classproperty
# from garoupa.uuid import UUID
# from tatu.abs.sql import SQL
#
#
# class SQLA(SQL):
#     def _config_(self):
#         return self._config
#
#     def __init__(self, engine):
#         self._config = locals().copy()
#         if not isinstance(engine, Engine):
#             raise Exception("Wrong engine:", type(engine))
#         del self._config["self"]
#         del self._config["__class__"]
#         print("str of engine:::::::::::", str(engine))
#         self._uuid = UUID((self.__class__.__name__ + str(engine)).encode())
#         self.engine = engine
#         self.hostname = socket.gethostname()
#         super().__init__(threaded=False, timeout=9999999, close_when_idle=False)
#
#     def _uuid_(self):
#         return self._uuid
#
#     def _open_(self):
#         """
#         Each reconnection has a cost of approximately 150ms in ADSL (ping=30ms).
#         :return:
#         """
#         if self.debug:
#             print("getting connection...")
#         self.connection = pymysql.connect(
#             host=self.host,
#             user=self.user,
#             password=self.password,
#             charset="utf8",
#             # cursorclass=pymysql.cursors.DictCursor,  # set at the bottom of this file
#             # client_flag=CLIENT.MULTI_STATEMENTS
#         )
#         self.connection.client_flag &= pymysql.constants.CLIENT.MULTI_STATEMENTS
#         self.connection.autocommit(False)
#         # self.connection.server_status
#
#         if self.debug:
#             print("getting cursor...")
#
#         # Create db if it doesn't exist yet.
#         with self.cursor() as c:
#             c.execute(*self.prepare(f"SHOW DATABASES LIKE '{self.db}'"))
#             setup = c.fetchone() is None
#             if setup:
#                 if self.debug:
#                     print("creating database", self.db, "on", self.database, "...")
#                 c.execute("create database if not exists " + self.db)
#                 self.commit()
#
#         if self.debug:
#             print("using database", self.db, "on", self.database, "...")
#         with self.cursor() as c:
#             c.execute("use " + self.db)
#             c.execute(f"show tables")
#
#         # Create tables if they don't exist yet.
#         try:
#             with self.cursor() as c:
#                 c.execute(f"select 1 from data")
#         except:
#             if self.debug:
#                 print("creating database", self.database, "...")
#             self._setup()
#             self.commit()
#
#         return self
#
#     @classproperty
#     def _now_function(cls):
#         return "now()"
#
#     @classproperty
#     def _keylimit(cls):
#         return "(190)"
#
#     @classproperty
#     def _auto_incr(cls):
#         return "AUTO_INCREMENT"
#
#     @classmethod
#     def _on_conflict(cls, cols):
#         return "ON DUPLICATE KEY UPDATE"
#
#     @classproperty
#     def _insert_ignore(cls):
#         return "insert ignore"
#
#     @classmethod
#     def _fkcheck(cls, enable):
#         return f"SET FOREIGN_KEY_CHECKS={'1' if enable else '0'};"
#
#     @classproperty
#     def _placeholder(cls):
#         return "%s"
#
#     def newcursor(self):
#         return self.connection.cursor(pymysql.cursors.DictCursor)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
import warnings
from unittest import TestCase

from aiuna.content.root import Root
from aiuna.step.dataset import Dataset
from aiuna.step.let import Let
from tatu import Tatu
from akangatu.transf.noop import NoOp


class TestTatu(TestCase):
    def setUp(self):
        warnings.simplefilter('ignore', (DeprecationWarning, UserWarning, ImportWarning))
        self.db = Tatu("sqlite://:memory:", threaded=False)
        d = Dataset().data
        d["X"] = [[1, 2, 3]]
        d["y"] = [1, 2, 3]
        self.d = d
        self.db.store(d)
        self.letstep = Let(field="X", value=[[1, 2, 3]])  # * Let(field="y", value=[1, 2, 3])

    def tearDown(self):
        self.db.close()

    def test__uuid_(self):
        self.assertEqual("01Dnf8SfdrCmMTWbGbwnoKu", self.db.id)

    def test_store__fetch__remove(self):
        d = Dataset().data
        self.db.store(d)
        self.assertEqual(d, self.db.fetch(d.id, lazy=False))
        # self.assertEqual(d, self.db.fetch(d.id, lazy=True))
        self.assertEqual(d, self.db.fetch(d))
        self.db.removedata(d)
        self.assertEqual(None, self.db.fetch(d))

    def test_fetchhistory(self):
        h = self.db.fetchhistory(self.d.id)
        self.assertEqual(self.d.history, h)

    def test_fetchstep(self):
        self.assertIsNone(self.db.fetchstep("xxxxxxxxx"))
        self.assertEqual(NoOp(), self.db.fetchstep(NoOp().id))
        self.assertEqual(self.letstep, self.db.fetchstep(self.letstep.id))

    def test_hasdata(self):
        self.assertFalse(self.db.hasdata("xxxxxxxxx"))
        self.assertFalse(self.db.hasdata(Root.id))
        self.assertTrue(self.db.hasdata(Root.id, include_empty=True))

    #
    def test_getdata(self):
        self.assertIsNone(self.db.getdata(Root.id, include_empty=False))
        self.assertEquals(Root.id, self.db.getdata(Root.id, include_empty=True)["parent"])
        # self.assertEquals(self.letstep, self.db.getdata((self.d >> self.letstep).id, include_empty=True)["parent"])

    # def test_hasstep(self):
    #     self.fail()
    #
    # def test_getstep(self):
    #     self.fail()
    #
    # def test_getfields(self):
    #     self.fail()
    #
    # def test_getcontent(self):
    #     self.fail()
    #
    # def test_hascontent(self):
    #     self.fail()
    #
    #
    # def test_lock(self):
    #     self.fail()
    #
    # def test_deldata(self):
    #     self.fail()
    #
    # def test_unlock(self):
    #     self.fail()
    #
    # def test_putdata(self):
    #     self.fail()
    #
    # def test_putcontent(self):
    #     self.fail()
    #
    # def test_putfields(self):
    #     self.fail()
    #
    # def test_storestep(self):
    #     self.fail()
    #
    # def test_putstep(self):
    #     self.fail()



# #  Copyright (c) 2020. Davi Pereira dos Santos
# #  This file is part of the tatu project.
# #  Please respect the license - more about this in the section (*) below.
# #
# #  tatu is free software: you can redistribute it and/or modify
# #  it under the terms of the GNU General Public License as published by
# #  the Free Software Foundation, either version 3 of the License, or
# #  (at your option) any later version.
# #
# #  tatu is distributed in the hope that it will be useful,
# #  but WITHOUT ANY WARRANTY; without even the implied warranty of
# #  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# #  GNU General Public License for more details.
# #
# #  You should have received a copy of the GNU General Public License
# #  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
# #
# #  (*) Removing authorship by any means, e.g. by distribution of derived
# #  works or verbatim, obfuscated, compiled or rewritten versions of any
# #  part of this work is a crime and is unethical regarding the effort and
# #  time spent here.
# #  Relevant employers or funding agencies will be notified accordingly.
# import json
# import warnings
# from unittest import TestCase
#
# from aiuna.compression import unpack, pack
# from aiuna.content.root import Root
# from aiuna.step.dataset import Dataset
# from aiuna.step.let import Let
# from app import create_app, db
# from app.config import Config
# from kururu.tool.enhancement.attribute.binarize import Binarize
#
# from tatu import Tatu
# from tatu.okast import OkaSt
# from tatu.sql.sqlite import SQLite
# from akangatu.transf.noop import NoOp
#
#
# class TestConfig(Config):
#     TESTING = True
#     SQLALCHEMY_DATABASE_URI = 'sqlite://'
#     TATU_URL = "sqlite://:memory:"
#
# class TestOkaSt(TestCase):
#     create_user1 = {
#         "username": "user1111",
#         "password": "password123",
#         "email": "teste1@teste.com",
#         "name": "Teste1"
#     }
#
#     def setUp(self):
#         warnings.simplefilter('ignore', (DeprecationWarning, UserWarning, ImportWarning))
#         self.app = create_app(TestConfig)
#         self.app_context = self.app.app_context()
#         self.app_context.push()
#         db.create_all()
#         self.app.config['JWT_BLACKLIST_ENABLED'] = False
#         self.iris = Dataset().data
#
#         # Create a storage identical to the one used in app, to check for its uuid in test__uuid_().
#         self.db = SQLite(db=":memory:")
#
#     def tearDown(self):
#         self.app.config["TATU_SERVER"].close()
#         self.db.close()
#         db.session.remove()
#         db.drop_all()
#         self.app_context.pop()
#
#     def create_user(self, c):
#         response = c.post("/api/users", json=self.create_user1)
#         self.assertEqual(response.status_code, 201)
#         return json.loads(response.get_data(as_text=True))
#
#     def get_token(self, c):
#         login1 = self.create_user1.copy()
#         del login1["email"]
#         del login1["name"]
#         response = c.post("/api/auth/login", json=login1)
#
#         self.assertEqual(200, response.status_code)
#         data = json.loads(response.get_data(as_text=True))
#         c.environ_base["HTTP_AUTHORIZATION"] = "Bearer " + data["access_token"]
#
#         return data["access_token"]
#
#     # def test__hasdata_(self):  # bool [ok not anymore, hascontent changed?]
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         o.store(self.iris, ignoredup=True)
#     #         self.assertFalse(o.hasdata("nonexistent uuid"))
#     #         self.assertFalse(o.hasdata(Root.id))
#     #         self.assertTrue(o.hasdata(Root.id, include_empty=True))
#     #         self.assertTrue(o.hasdata(self.iris.id))
#
#     def test__getdata_(self):  # None or jsonable dict [ok]
#         with self.app.test_client() as c:
#             self.create_user(c)
#             o = OkaSt(self.get_token(c), url=c)
#             self.assertIsNone(o.getdata(Root.id, include_empty=False))
#             self.assertEquals(Root.id, o.getdata(Root.id, include_empty=True)["parent"])
#
#     def test__uuid_(self):  # str [ok]
#         with self.app.test_client() as c:
#             self.create_user(c)
#             o = OkaSt(self.get_token(c), url=c)
#             self.assertEqual(o.id, self.db.id)
#
#     def test__hasstep_(self):  # bool [ok]
#         with self.app.test_client() as c:
#             self.create_user(c)
#             o = OkaSt(self.get_token(c), url=c)
#             self.assertFalse(o.hasstep("nonexistent uuid"))
#             self.assertTrue(o.hasstep(NoOp().id))
#
#     def test__getstep_(self):  # None or jsonable dict [ok]
#         with self.app.test_client() as c:
#             self.create_user(c)
#             o = OkaSt(self.get_token(c), url=c)
#             self.assertIsNone(o.getstep("nonexistent uuid"))
#             self.assertEquals(NoOp().context, o.getstep(NoOp().id)["path"])
#
#     # def test__getfields_(self):  # None or dict of binaries []
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         self.assertIsNone(o.getfields("nonexistent uuid"))
#     #         self.assertDictEqual(o.getfields(Root.id), {})
#     #         self.assertEquals({k: pack(v) for k, v in self.iris.items()}, o.getfields(self.iris.id))
#
#     # def test__hascontent_(self):  # entra e sai lista de str-uuids []
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         self.assertListEqual(o.hascontent(["nonexistent uuid"]), [])
#     #         self.assertListEqual(o.hascontent([self.iris.uuids["X"].id]), [self.iris.uuids["X"].id])
#
#     # def test__getcontent_(self):  # None or binary [ok]
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         o.store(self.iris, ignoredup=True)
#     #         self.assertFalse(o.getcontent("nonexistent uuid"))
#     #         self.assertEquals(pack(self.iris.X), o.getcontent(self.iris.uuids["X"].id))
#
#     # def test__lock_(self):  # bool []
#     #     iris2 = self.iris >> Let("F", [1, 2])
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         # self.assertFalse(o.lock(self.iris.id, check_existence=False))
#     #         self.assertTrue(o.lock(iris2.id))
#
#     # def test__unlock_(self):  # bool []
#     #     iris2 = self.iris >> Let("F", [1, 2])
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         # self.assertFalse(o.unlock(self.iris.id))
#     #         self.assertTrue(o.unlock(iris2.id))
#
#     # def test__putcontent_(self):  # bool: send binary []
#     #     iris3 = self.iris >> Let("F", [1, 2, 3])
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         self.assertFalse(o.putcontent(iris3.uuids["X"].id, iris3.X))
#     #         self.assertTrue((unpack(o.getcontent(iris3.uuids["X"].id)) == iris3.X).all)
#     #         print()
#     #         print("-------------------------------")
#     #         # self.assertTrue(o.putcontent(iris3.uuids["F"].id, iris3.F))
#     #         # self.assertTrue((unpack(o.getcontent(iris3.uuids["F"].id)) == iris3.F).all)
#
#     # def test__putdata_(self):  # bool:  send jsonable dict []
#     #     iris4 = self.iris >> Let("F", [1, 2, 3, 4])
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         dic = {
#     #             "id": iris4.id,
#     #             "step": iris4.step_uuid.id,
#     #             "inn": None,
#     #             "stream": False,
#     #             "parent": "nonexistent uuid",
#     #             "locked": False
#     #         }
#     #         self.assertFalse(o.putdata(**dic))
#     #
#     #         dic["parent"] = Root.id
#     #         self.assertTrue(o.putdata(**dic))
#
#     # def test__putfields_(self):  # bool: send jsonable dict []
#     #     iris5 = self.iris >> Let("F", [1, 2, 3, 4, 5])
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         rows = {"rows": [
#     #             {"data": iris5.id, "name": "X", "content": iris5.uuids["X"].id},
#     #             {"data": iris5.id, "name": "Y", "content": iris5.uuids["Y"].id}
#     #         ]}
#     #
#     #         self.assertFalse(o.putfields(rows))
#     #         self.assertTrue(o.putcontent(iris5.uuids["X"], pack(iris5.X)))
#     #         self.assertTrue(o.putcontent(iris5.uuids["Y"], pack(iris5.Y)))
#     #         self.assertTrue(o.putcontent(iris5.uuids["F"], pack(iris5.F)))
#     #         self.assertTrue(o.putfields(rows))
#
#     # def test__putstep_(self):  # bool: send jsonable dict []
#     #     with self.app.test_client() as c:
#     #         self.create_user(c)
#     #         o = OkaSt(self.get_token(c), url=c)
#     #         dic = {
#     #             "id": NoOp().id,
#     #             "name": NoOp().name,
#     #             "path": NoOp().context,
#     #             "config": {}
#     #         }
#     #         self.assertFalse(o.putstep(**dic))
#     #
#     #         dic = {
#     #             "id": Binarize().id,
#     #             "name": Binarize().name,
#     #             "path": Binarize().context,
#     #             "config": {}
#     #         }
#     #         self.assertTrue(o.putstep(**dic))



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import socket
import sqlite3

from garoupa.decorator import classproperty
from garoupa.uuid import UUID
from tatu.abs.sql import SQL


class SQLite(SQL):
    def _config_(self):
        return self._config

    def __init__(self, db="tatu-sqlite", threaded=True, close_when_idle=False, storage_info=None, debug=False,
                 read_only=False):
        self._config = locals().copy()
        del self._config["self"]
        del self._config["__class__"]
        self.info = db
        self.read_only = read_only
        self.hostname = socket.gethostname()
        self.database = db if ":memory:" in db else (db + ".db")
        self.storage_info = storage_info
        self.debug = debug
        self._uuid = UUID((self.__class__.__name__ + db).encode())
        super().__init__(threaded, timeout=2, close_when_idle=close_when_idle)

    def _uuid_(self):
        return self._uuid

    def _open_(self):
        # isolation_level=None -> SQLite autocommiting
        # isolation_level='DEFERRED' -> SQLite transactioning
        self.connection = sqlite3.connect(self.database, isolation_level='DEFERRED')
        self.connection.row_factory = sqlite3.Row

        # Create tables if they don't exist yet.
        try:
            with self.cursor() as c:
                c.execute(f"select 1 from data")
        except:
            if self.debug:
                print("creating database", self.database, "...")
            self._setup()
            self.commit()

    @classproperty
    def _now_function(cls):
        return "datetime()"

    @classproperty
    def _keylimit(cls):
        return ""

    @classproperty
    def _auto_incr(cls):
        return "AUTOINCREMENT"

    @classmethod
    def _on_conflict(cls, cols):
        return f"ON CONFLICT{cols} DO UPDATE SET"

    @classmethod
    def _fkcheck(cls, enable):
        return f"PRAGMA foreign_keys = {'ON' if enable else 'OFF'};"

    @classproperty
    def _insert_ignore(cls):
        return "insert or ignore"

    @classproperty
    def _placeholder(cls):
        return "?"

    def newcursor(self):
        return self.connection.cursor()



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from typing import Optional, List

from aiuna.content.data import Data
from garoupa.uuid import UUID
from tatu.storageinterface import StorageInterface


class Amnesia(StorageInterface):
    def _config_(self):
        return self._config

    def _fetch_children_(self, data: Data):
        raise Exception("(Pseudo)Storage Amnesia cannot retrieve children!")

    def __init__(self):
        self._config = {}
        super().__init__(threaded=False, timeout=None, close_when_idle=False)

    def _uuid_(self):
        return UUID(b"Amnesia")

    def _fetch_(self, data: Data, lock=False) -> Optional[Data]:
        return None

    def _delete_(self, data: Data, check_missing=True):
        pass

    def _store_(self, data: Data, check_dup=True):
        pass

    def _unlock_(self, data):
        pass



#  Copyright (c) 2020. Davi Pereira dos Santos and Rafael Amatte Bisão
#  This file is part of the tatu project.
#  Please respect the license - more about this in the section (*) below.
#
#  tatu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  tatu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with tatu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
from typing import Union

import requests as req

from garoupa.uuid import UUID
from tatu.storageinterface import StorageInterface

default_url = 'http://data.analytics.icmc.usp.br'


def j(r):
    """Helper function needed because flask test_client() provide json as a property(?), not as a method."""
    return r.json() if callable(r.json) else r.json


class OkaSt(StorageInterface):
    """Central remote storage"""

    def _config_(self):
        return self._config

    def __init__(self, token="Invalid",
                 alias=None, threaded=True, url: Union[callable, str] = "http://localhost:5000", close_when_idle=False):
        self._config = locals().copy()
        del self._config["self"]
        del self._config["__class__"]
        # print("STORAGE: ", url)
        self.token = token
        self.external_requests = url if callable(url) else self.request
        self.url = url
        self.alias = alias
        self.prefix = self.url if isinstance(self.url, str) else ""
        # TODO: check if threading will destroy oka
        super().__init__(threaded, timeout=6, close_when_idle=close_when_idle)

    def request(self, route, method, **kwargs):
        headers = {'Authorization': 'Bearer ' + self.token} if self.token else {}
        r = getattr(req, method)(self.url + route, headers=headers, **kwargs)
        if r.status_code == 401:
            print("Please login before.")
            from tatu.auth import gettoken
            self.token = gettoken(self.url)
            return self.request(route, method, **kwargs)
        elif r.status_code == 422:
            # msg = j(r)["errors"]["json"]
            # print(j(r))
            pass
        else:
            if r.ok:
                return r
            print(r.content)
            print(j(r))
            print(j(r)["errors"])
            msg = j(r)["errors"]["json"]
            print(msg)
            raise Exception(msg)
    
    def _uuid_(self):
        # REMINDER syncing needs to know the underlying storage of okast, because the token is not constant as an identity
        return UUID(j(self.request(f"/api/sync_uuid", "get"))["uuid"])

    def _hasdata_(self, id, include_empty):
        url = f"/api/sync?uuids={id}&cat=data&fetch=false&empty={include_empty}"
        return j(self.request(url, "get"))["has"]

    def _hasstream_(self, data):
        url = f"/api/sync?uuids={data.id}&cat=stream&fetch=false"
        return j(self.request(url, "get"))["has"]

    def _getdata_(self, id, include_empty):
        url = f"/api/sync?uuids={id}&cat=data&fetch=true&empty={include_empty}"
        return j(self.request(url, "get"))

    def _getstream_(self, data):
        url = f"/api/sync?uuids={data}&cat=stream&fetch=true"
        return j(self.request(url, "get"))

    def _hasstep_(self, id):
        url = f"/api/sync?uuids={id}&cat=step&fetch=false"
        return j(self.request(url, "get"))["has"]

    def _getstep_(self, id):
        url = f"/api/sync?uuids={id}&cat=step&fetch=true"
        return j(self.request(url, "get"))

    def _getfields_(self, id):
        url = f"/api/sync/{id}/many&cat=fields"
        return j(self.request(url, "get"))

    def _hascontent_(self, ids):
        uuids = "&".join([f"uuids={id}" for id in ids])
        url = f"/api/sync?{uuids}&cat=content&fetch=false"
        return j(self.request(url, "get"))["has"]

    def _getcontent_(self, id):
        url = f"/api/sync/{id}/content"
        r = self.request(url, "get")
        return None if r.content == b'null\n' else r.content

    def _lock_(self, id):
        url = f"/api/sync/{id}/lock"
        return j(self.request(url, "put"))["success"]

    def _unlock_(self, id):
        url = f"/api/sync/{id}/unlock"
        return j(self.request(url, "put"))["success"]

    def _putdata_(self, id, step, inn, stream, parent, locked, ignoredup):
        kwargs = locals().copy()
        del kwargs["self"]
        url = f"/api/sync?cat=data"
        return j(self.request(url, "post", json={"kwargs": kwargs}))["success"]

    def _putstream_(self, rows, ignoredup):
        url = f"/api/sync/many?cat=stream&ignoredup={ignoredup}"
        return j(self.request(url, "post", json={"rows": rows}))["n"]

    def _putfields_(self, rows, ignoredup):
        url = f"/api/sync/many?cat=fields&ignoredup={ignoredup}"
        return j(self.request(url, "post", json={"rows": rows}))["n"]

    def _putcontent_(self, id, value, ignoredup):
        url = f"/api/sync/{id}/content?ignoredup={ignoredup}"
        return j(self.request(url, "post", files={'bina': value}))["success"]

    def _putstep_(self, id, name, path, config, dump, ignoredup):
        kwargs = locals().copy()
        del kwargs["self"]
        url = f"/api/sync?cat=step"
        return j(self.request(url, "post", json={"kwargs": kwargs}))["success"]

    def _deldata_(self, id):
        raise Exception(f"OkaSt cannot delete Data entries! HINT: deactivate post {id} on Oka.")

    def _open_(self):
        pass  # nothing to open for okast

    def _close_(self):
        pass  # nothing to close for okast

# TODO: consultar previamente o que falta enviar, p/ minimizar trafego
#     #  TODO: enviar por field
#     #  TODO: override store() para evitar travessia na classe mãe?



# from unittest import TestCase
#
# from aiuna.step.dataset import dataset
# from tatu import Tatu
# from flask import Flask
#
# from flask_sqlalchemy import SQLAlchemy
#
# class TestSQLA(TestCase):
#     def insert(self):
#         app = Flask(__name__)
#         db = SQLAlchemy(app)
#         db =??
#         tatu = Tatu(db)
#
#         d = dataset.data
